﻿/*
//  Реализация алгоритма последовательного умножения 
//  двух квадратных матриц A и B.
//
//  API
//  int SerialMatrixMultiplication(
//                                 double* pAMatrix, double* pBMatrix,
//                                 double* pCMatrix, const int Size);   
//
//  Input:
//     + pAMatrix - указатель на вектор памяти элементов типа double,
//	                матрица А, i-я строка начинается с Size*i индекса 
//     + pBMatrix - указатель на вектор памяти элементов типа double,
//	                матрица B, i-я строка начинается с Size*i индекса 
//     + Size     - переменная типа int, размерность квадратных матриц
//  Output:
//     + pCMatrix - указатель на вектор памяти элементов типа double,
//	                матрица C, i-я строка начинается с Size*i индекса
//  Result:
//     + функция возвращает значение типа int, код завершения функции
*/
int SerialMatrixMultiplication(
                               double* pAMatrix, double* pBMatrix,
                               double* pCMatrix, const int Size)
{
    int Res = PARLIB_SUCCESS;
    int i, j, k;

    assert(pAMatrix);
    assert(pBMatrix);
    assert(pCMatrix);
    assert(Size > 0);

    for( i = 0; i < Size * Size; ++i )
    {
        pCMatrix[i] = 0.0;
    }

    // Последовательное перемножение матриц А и В
    // i-я строка матрицы А скалярно умножается на 
    // j-ый столбец матрицы В, получяется 
    // элемент матрицы Cij
    for( i = 0; i < Size; ++i )
    {
        for( j = 0; j < Size; ++j )
        {
            for( k = 0; k < Size; ++k )
            {
                pCMatrix[i*Size+j] += pAMatrix[i*Size+k] * pBMatrix[k*Size+j];
            }
        }
    }
    return Res;
}