﻿/*
//  Реализация базового алгоритма умножения 
//  квадратной матрицы A на вектор b с использованием технологии OpenMP.
// int OMP_baseMatrixVectorMultiplication(const double* pMatrix,
//                     const double* pVector, double* pResult, const int Size);
//
//  В основу этого алгоритма умножения матрицы на вектор положено
//  разбиение матрицы A на непрерывные последовательности строк
//  (горизонтальные полосы).
//
//  Для организации параллельных вычислений, в качестве базовой подзадачи
//  выбирается вычисление одного элемента вектора с в одном потоке.
//
//  Input:
//
//      + const double* pMatrix - исходная матрица;
//      + const double* pVector - исходный вектор;
//      + const int Size - размер исходной квадратной матрицы;
//
//  Output:
//
//      + double* pResult - результат перемножения;
//
//  Result:
//
//      + функция возвращает значение типа int - код завершения функции
*/

int OMP_baseMatrixVectorMultiplication(const double* pMatrix,const double* pVector, double* pResult,const int Size)
{
  int i, j;
  assert(pMatrix);
  assert(pVector);
  assert(pResult);
  assert(Size>0);
#pragma omp parallel for private (j)  
  for (i=0; i<Size; i++)
  {
    for (j=0; j<Size; j++)
    {
      pResult[i] += pMatrix[i*Size+j]*pVector[j];
    }
  } //for (i=0; i<Size; i++)
  return 0;
}