﻿/*
// Реализация алгоритма последовательной сортировки 
// массива "пузырьком" 
//
// API
// int SerialBubbleSort(double* pData, const int Size);
//
//  Input:
//
//     +  double* pData - исходный массив;
//     +  const int Size - размер массива;
//
//  Output:
//
//     +  double* pData - массив, отсортированный в порядке возрастания;
//
//  Result:
//     + функция возвращает значение типа int - код завершения функции.
*/

int SerialBubbleSort(double* pData, const int Size)
{
  double temp;
  int i, j;
  int f = 0;
  assert(pData);
  assert(Size>0);
  for (i=0; (i<Size-1) && (f==0); i++) 
  {
    f = 1;
    for (j=0; j<Size-1-i; j++)
    {
      // переставление элементов местами
      if ( pData[j] > pData[j+1] ) 
      {
        f=0;
        temp = pData[j];
        pData[j] = pData[j+1];
        pData[j+1] = temp;
      }
    }//for (int j=0; j<Size-1-i; j++)
  }//for (int i=0; i<Size-1; i++) 
  return 0;
}


/*
// Реализация алгоритма последовательной сортировки 
// массива методом чётной-нечётной перестановки. 
//
// API
// int int OddEvenSort ( double* pData, const int Size );
//
//  Input:
//
//     +  double* pData - исходный массив;
//     +  const int Size - размер массива;
//
//  Output:
//
//     +  double* pData - массив, отсортированный в порядке возрастания;
//
//  Result:
//     + функция возвращает значение типа int - код завершения функции.
*/

int OddEvenSort ( double* pData, const int Size )
{
  double temp;
  int upper_bound;
  int i, j;
  assert(pData);
  assert(Size>0);
  if (Size%2==0)
  {
    upper_bound = Size/2-1;
  }
  else
  {
    upper_bound = Size/2;
  }

  for (i=0; i<Size; i++) 
  {
    // чётная итерация
    if (i%2 == 0) 
    {
      for (j=0; j<Size/2; j++)
      {
        if(pData[2*j] > pData[2*j+1])
        {
          temp = pData[2*j];
          pData[2*j] = pData[2*j+1];
          pData[2*j+1] = temp;
        }
      }// for (int j=0; j<Size/2; j++)
    }// if (i%2 == 0)  
    //нечётная итерация
    else    
    {     
      for (j=0; j<upper_bound; j++)
      {
        if(pData[2*j+1] > pData[2*j+2])
        {
          temp = pData[2*j+1];
          pData[2*j+1] = pData[2*j+2];
          pData[2*j+2] = temp;
        }
      }//for (int j=0; j<upper_bound; j++)
    }// else
  }// for (int i=0; i<Size; i++)
  return 0;
}


/*
// Реализация алгоритма обмена местами двух элементов одного массива 
//
// API
// void swap(double* pData, int i, int j) 
//
//  Input:
//
//     +  double* pData - исходный массив;
//     +  int i - номер первого элемента;
//     +  int i - номер второго элемента;
//
//  Output:
//
//     +  double* pData - массив, отсортированный в порядке возрастания;
//
//  Result:
//     + функция возвращает значение типа int - код завершения функции.
*/

void swap(double* pData, int i, int j)
{
  double temp;
  assert(pData);
  assert(i>=0);
  assert(j>=0);
  temp = pData[i];
  pData[i] = pData[j];
  pData[j] = temp;
}

/*
// Реализация алгоритма последовательной сортировки 
// массива методом Шелла. 
//
// API
// int int ShellSort ( double* pData, const int Size );
//
//  Input:
//
//     +  double* pData - исходный массив;
//     +  const int Size - размер массива;
//
//  Output:
//
//     +  double* pData - массив, отсортированный в порядке возрастания;
//
//  Result:
//     + функция возвращает значение типа int - код завершения функции.
*/

int ShellSort(double* A, const int n)
{
  int incr = n/2;
  int i,j;
  assert(A);
  assert(n>0);
  while( incr > 0 )
  {
    for (i=incr; i<n; i++  )
    {
      j = i-incr; 
      while ( j >= 0 )
        if ( A[j] > A[j+incr] )
        { 
          swap(A, j,j+incr);
          j = j - incr;   
        }
        else j = -1;
    }
    incr = incr/2; 
  }
  return 0;
}

/*
// Реализация алгоритма последовательной сортировки 
// части массива отрезка [from, to] методом быстрой сортировки Хоара. 
//
// API
// void QuickSorter ( double A[],  int from, int to );
//
//  Input:
//
//     +  double A[] - исходный массив;
//     +  int from - начальный индекс сортировки;
//     +  int to - конечный индекс сортировки;
//
//  Output:
//
//     +  double A[] - часть массива, отсортированный в порядке возрастания;
*/

void QuickSorter ( double A[],  int from, int to )
{
  int  i, j;
  double x, temp;
  assert(A);
  assert(from>=0);
  if ( from >= to ) return;
  i = from;
  j = to;
  x = A[ ( from + to ) / 2];
  while ( i <= j )
  {
    while ( A[i] < x )
    {
      i ++;
    }
    while ( A[j] > x ) 
    {
      j --;
    }
    if ( i <= j ) 
    {
      temp = A[i]; A[i] = A[j]; A[j] = temp;
      i ++;
      j --;
    }
  }
  QuickSorter ( A, from, j );
  QuickSorter ( A, i, to );
}

/*
// Реализация алгоритма последовательной сортировки 
// массива методом быстрой сортировки Хоара. 
//
// API
// int SerialQuickSort ( double* pData, const int Size );
//
//  Input:
//
//     +  double* pData - исходный массив;
//     +  const int Size - размер массива;
//
//  Output:
//
//     +  double* pData - массив, отсортированный в порядке возрастания;
//
//  Result:
//     + функция возвращает значение типа int - код завершения функции.
*/

int SerialQuickSort(double *pData, const int Size)
{
  assert(pData);
  assert(Size>0);
  QuickSorter(pData, 0, Size-1);
  return 0;
}

int merge(double *first1, double *last1, double *first2, double *last2, double *result)
{
  while(1)
  {
    *result++ = ((*first2)<(*first1))? *first2++ : *first1++;
    if (first1 == last1) 
    {
      return copy(first2, last2, result);
    }
    if (first2 == last2) 
    {
      return copy(first1, last1, result);
    }
  }
  return 0;
}

void MergeBlock(double *pBlock, double *pRecvBlock, double *pDoubleBlock, const int BlockSize)
{
  int ProcRank;
  int i;
  double *pFirst;
  double *pSecond;
  MPI_Comm_rank(MPI_COMM_WORLD, &ProcRank);
  pFirst = pBlock;
  pSecond = pRecvBlock;
  for(i = 0; i < 2 * BlockSize; i++)
  {
    if ( ((*pFirst) <= (*pSecond))  && (pFirst != (pBlock + BlockSize)) )
    {
      pDoubleBlock[i] = (*pFirst);
      pFirst++;
    }
    else
    {
      if(pSecond != pRecvBlock + BlockSize)
      {
        pDoubleBlock[i] = (*pSecond);
        pSecond++;
      }
      else
      {
        pDoubleBlock[i] = (*pFirst);
        pFirst++;
      }
    }// else
  }// for(int i = 0; i < 2 * BlockSize; i++)
}

int MPI_Odd_even_sort(double *pData, const int Size)
{
  MPI_Status Status;
  int ProcNum, ProcRank;
  int BlockSize;
  int i, j;
  double *pBlock;
  double *pRecvBlock;
  double *pDoubleBlock;

  MPI_Comm_size(MPI_COMM_WORLD, &ProcNum);
  MPI_Comm_rank(MPI_COMM_WORLD, &ProcRank);
  if(ProcNum == 1)
  {
    SerialQuickSort(pData, Size);
    return 0;
  }
  BlockSize = Size / ProcNum;
  if(BlockSize == 0)
  {
    BlockSize = 1;
    ProcNum = Size;
  }

  assert (Size % ProcNum == 0);
  pBlock = (double*) malloc ( BlockSize * sizeof(double));
  pRecvBlock = (double*) malloc ( BlockSize * sizeof(double));
  pDoubleBlock = (double*) malloc ( 2 * BlockSize * sizeof(double) );

  MPI_Scatter(pData, BlockSize, MPI_DOUBLE, pBlock, BlockSize, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  SerialQuickSort(pBlock, BlockSize);

  for(i = 0; i < ProcNum; i++)
  {
    // чётная итерация 
    if(i % 2 == 0)
    {
      if( ProcRank % 2 == 1)
      {
        if ((ProcRank != ProcNum - 1) || (ProcNum % 2 == 0))
        {
          MPI_Recv(pRecvBlock, BlockSize, MPI_DOUBLE, ProcRank - 1, 0, MPI_COMM_WORLD, &Status);         
          //MergeBlock(pBlock, pRecvBlock, pDoubleBlock, BlockSize);
          merge(pBlock, pBlock + BlockSize, pRecvBlock, pRecvBlock + BlockSize, pDoubleBlock);
          MPI_Send( pDoubleBlock, BlockSize, MPI_DOUBLE, ProcRank - 1, 0, MPI_COMM_WORLD);
          for(j = 0; j < BlockSize; j++)
          {
            pBlock[j] = pDoubleBlock[j + BlockSize];
          }
        }
      }
      else
      {
        if (ProcRank != ProcNum - 1)
        {
          MPI_Send(pBlock, BlockSize, MPI_DOUBLE, ProcRank + 1, 0, MPI_COMM_WORLD);
          MPI_Recv(pBlock, BlockSize, MPI_DOUBLE, ProcRank + 1, 0, MPI_COMM_WORLD, &Status);
        }
      } // else
    }// if(i % 2 == 0) 
    else
    {
      if((ProcRank != 0))//&&(ProcRank != ProcNum - 1))
      {
        if ( ProcRank % 2 == 0)
        {
          MPI_Recv(pRecvBlock, BlockSize, MPI_DOUBLE, ProcRank - 1, 0, MPI_COMM_WORLD, &Status); 
          //MergeBlock(pBlock, pRecvBlock, pDoubleBlock, BlockSize);
          merge(pBlock, pBlock + BlockSize, pRecvBlock, pRecvBlock + BlockSize, pDoubleBlock);
          MPI_Send( pDoubleBlock, BlockSize, MPI_DOUBLE, ProcRank - 1, 0, MPI_COMM_WORLD);
          for(j = 0; j < BlockSize; j++)
          {
            pBlock[j] = pDoubleBlock[j + BlockSize];
          }
        }
        else
        {
          if (ProcNum % 2 == 1)
          {
            MPI_Send(pBlock, BlockSize, MPI_DOUBLE, ProcRank + 1, 0, MPI_COMM_WORLD);
            MPI_Recv(pBlock, BlockSize, MPI_DOUBLE, ProcRank + 1, 0, MPI_COMM_WORLD, &Status);
          }
          else
          {
            if(ProcRank != ProcNum - 1)
            {
              MPI_Send(pBlock, BlockSize, MPI_DOUBLE, ProcRank + 1, 0, MPI_COMM_WORLD);
              MPI_Recv(pBlock, BlockSize, MPI_DOUBLE, ProcRank + 1, 0, MPI_COMM_WORLD, &Status);
            }
          }
        } // else
      }// if((ProcRank != 0)&&(ProcRank != ProcNum - 1))

    }//else

  }// for

  MPI_Gather(pBlock, BlockSize, MPI_DOUBLE, pData, BlockSize, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  free ( pBlock );
  free ( pRecvBlock );
  free ( pDoubleBlock );
  return 0; 

}


void DataDistributionSort(double *pData, int Size, double *pProcData, int BlockSize) { 

  // Allocate memory for temporary objects
  int ProcNum, ProcRank;
  int *pSendInd;
  int *pSendNum;
  int RestData;
  int CurrentSize;
  int i;
  MPI_Comm_size(MPI_COMM_WORLD, &ProcNum);
  MPI_Comm_rank(MPI_COMM_WORLD, &ProcRank);
  pSendInd = (int*) malloc ( ProcNum * sizeof(int));
  pSendNum = (int*) malloc ( ProcNum * sizeof(int));

  RestData = Size;
  CurrentSize  = Size / ProcNum;
  pSendNum[0] = CurrentSize ;
  pSendInd[0] = 0;
  for(i = 1; i < ProcNum; i++) 
  {
    RestData    -= CurrentSize;
    CurrentSize  = RestData / (ProcNum - i);
    pSendNum[i]  = CurrentSize;
    pSendInd[i]  = pSendInd[i - 1] + pSendNum[i - 1];
  }

  MPI_Scatterv(pData, pSendNum, pSendInd, MPI_DOUBLE, pProcData, 
    pSendNum[ProcRank], MPI_DOUBLE, 0, MPI_COMM_WORLD);

  free ( pSendNum );
  free ( pSendInd );                
}

void DataCollection(double *pData, int Size, double *pProcData, int BlockSize) { 

  // Allocate memory for temporary objects
  int ProcNum, ProcRank;
  int *pReceiveNum;
  int *pReceiveInd;
  int RestData;
  int i;
  MPI_Comm_size(MPI_COMM_WORLD, &ProcNum);
  MPI_Comm_rank(MPI_COMM_WORLD, &ProcRank);
  pReceiveNum = (int*) malloc ( ProcNum * sizeof(int));
  pReceiveInd = (int*) malloc ( ProcNum * sizeof(int));

  RestData = Size;
  pReceiveInd[0] = 0;
  pReceiveNum[0] = Size / ProcNum;
  for(i = 1; i < ProcNum; i++) 
  {
    RestData -= pReceiveNum[i - 1];
    pReceiveNum[i] = RestData / (ProcNum - i);
    pReceiveInd[i] = pReceiveInd[i - 1] + pReceiveNum[i - 1];
  }

  MPI_Gatherv(pProcData, BlockSize, MPI_DOUBLE, pData, 
    pReceiveNum, pReceiveInd, MPI_DOUBLE, 0, MPI_COMM_WORLD);

  // Free the memory
  free ( pReceiveNum ); 
  free ( pReceiveInd );
}

void ExchangeDataSort(double *pProcData, int BlockSize, int DualRank, 
                  double *pDualData, int DualBlockSize) 
{

  MPI_Status status;
  MPI_Sendrecv(pProcData, BlockSize, MPI_DOUBLE, DualRank, 0,
    pDualData, DualBlockSize, MPI_DOUBLE, DualRank, 0,
    MPI_COMM_WORLD, &status);
}


int copy(double *Begin, double *End, double *Result)
{

  while (Begin != End)
  {
    (*Result) = (*Begin);
    Begin++;
    Result++;
  }
  return 0;
}
//int merge(double *first1, double *last1, double *first2, double *last2, double *result)
//{
//  while(1)
//  {
//    *result++ = ((*first2)<(*first1))? *first2++ : *first1++;
//    if (first1 == last1) 
//    {
//      return copy(first2, last2, result);
//    }
//    if (first2 == last2) 
//    {
//      return copy(first1, last1, result);
//    }
//  }
//  return 0;
//}
void ParallelBubble(double *pProcData, int BlockSize) 
{
  //enum split_mode { KeepFirstHalf, KeepSecondHalf };
  // Local sorting the process data
  int ProcNum, ProcRank;
  int Offset;
  int i;
  int DualBlockSize;
  MPI_Status status;
  double *pDualData;
  double *pMergedData;
  int SplitMode = 0;
  MPI_Comm_size(MPI_COMM_WORLD, &ProcNum);
  MPI_Comm_rank(MPI_COMM_WORLD, &ProcRank);
  SerialBubbleSort(pProcData, BlockSize); 




  for(i = 0; i < ProcNum; i++) 
  { 
    if((i % 2) == 1) 
    { 
      if((ProcRank % 2) == 1) 
      { 
        Offset    = 1;
        SplitMode = 0;
      }
      else 
      {
        Offset    = -1;
        SplitMode = 1;
      }
    }
    else 
    { 
      if((ProcRank % 2) == 1) 
      { 
        Offset    = -1;
        SplitMode = 1;
      }
      else 
      { 
        Offset    = 1;
        SplitMode = 0;
      }
    }

    // Check the first and last processes
    if((ProcRank == ProcNum - 1) && (Offset ==  1)) 
    {
      continue;
    }
    if((ProcRank == 0 ) && (Offset == -1)) 
    {
      continue;
    }

    MPI_Sendrecv(&BlockSize, 1, MPI_INT, ProcRank + Offset, 0, 
      &DualBlockSize, 1, MPI_INT, ProcRank + Offset, 0,
      MPI_COMM_WORLD, &status);

    pDualData   = (double*) malloc ( DualBlockSize * sizeof(double));
    pMergedData = (double*) malloc ( (BlockSize + DualBlockSize) * sizeof(double) );

    // Data exchange
    ExchangeDataSort(pProcData, BlockSize, ProcRank + Offset, pDualData, DualBlockSize);

    // Data merging
    merge(pProcData, pProcData + BlockSize, pDualData, pDualData + DualBlockSize, pMergedData);

    // Data splitting
    if(SplitMode == 0)
    {
      copy(pMergedData, pMergedData + BlockSize, pProcData);
    }
    else
    {
      copy(pMergedData + BlockSize, pMergedData + BlockSize + DualBlockSize, pProcData);
    }

    free ( pDualData );
    free (pMergedData );
  }
}

/*
// Реализация алгоритма параллельной сортировки 
// массива "пузырьком". 
//
// API
// int MPI_Bubble_sort ( double* pData, const int Size );
//
//  Input:
//
//     +  double* pData - исходный массив;
//     +  const int Size - размер массива;
//
//  Output:
//
//     +  double* pData - массив, отсортированный в порядке возрастания;
//
//  Result:
//     + функция возвращает значение типа int - код завершения функции.
*/


int MPI_Bubble_sort ( double *pData, const int Size)
{
  double *pProcData = 0;
  //  int Size = 0;
  int BlockSize = 0;
  int i;
  int ProcRank, ProcNum;
  // double *pSerialData = 0;
  int RestData = Size; 
  MPI_Comm_size(MPI_COMM_WORLD, &ProcNum);
  MPI_Comm_rank(MPI_COMM_WORLD, &ProcRank);
  for( i = 0; i < ProcRank; i++)
  {
    RestData -= RestData / (ProcNum - i);
  }
  BlockSize = RestData / (ProcNum - ProcRank);

  pProcData = (double*) malloc ( BlockSize * sizeof(double) );

  // Distributing the initial data between processes
  DataDistributionSort(pData, Size, pProcData, BlockSize);

  // Testing the distribution
  //TestDistribution(pData, Size, pProcData, BlockSize);

  // Parallel bubble sort
  ParallelBubble(pProcData, BlockSize);
  // Print the sorted data
  //ParallelPrintData(pProcData, BlockSize);

  // Execution of data collection
  DataCollection(pData, Size, pProcData, BlockSize);
  return 0;
}