﻿/*
  Реализация базового алгоритма умножения 
  двух квадратных матриц A и B, при котором строка матрицы А 
  умножается на столбец матрицы В.

  API
  int MPI_BaseMatrixMultiplication(
                                 double pAMatrix, double pBMatrix,
                                 double pCMatrix, const int Size);   

  Input
     + pAMatrix - указатель на вектор памяти элементов типа double,
	                матрица А, i-я строка начинается с Sizei индекса 
     + pBMatrix - указатель на вектор памяти элементов типа double,
	                матрица B, i-я строка начинается с Sizei индекса 
     + Size     - переменная типа int, размерность квадратных матриц
  Output
     + pCMatrix - указатель на вектор памяти элементов типа double,
	                матрица C, i-я строка начинается с Sizei индекса
  Result
     + функция возвращает значение типа int, код завершения функции
*/

int MPI_BaseMatrixMultiplication( double pAMatrix, double pBMatrix,
                                double pCMatrix, const int Size )
{
 
  MPI_Datatype ColumnType, Rtp;
  int ProcNum, ProcRank;
  MPI_Status status;
  int StripSize;
  double Result;
  double rbuf;
  double vertStrip;
  int i, j, z, k;
  MPI_Comm_size ( MPI_COMM_WORLD, &ProcNum);
  MPI_Comm_rank ( MPI_COMM_WORLD, &ProcRank);
  StripSize = Size  ProcNum;

  if (StripSize == 0)
  {
    StripSize = 1;
  }
  Result = (double) malloc (StripSize  StripSize  sizeof(double));  блок результата

  rbuf = (double) malloc (Size  StripSize  sizeof(double));  горизонтальная полоса
  vertStrip =(double) malloc (Size  sizeof(double));  столбец
  MPI_Type_vector ( Size, 1, Size, MPI_DOUBLE, &ColumnType );
  MPI_Type_commit ( &ColumnType );
  MPI_Type_vector ( StripSize, StripSize, Size, MPI_DOUBLE, &Rtp );
  MPI_Type_commit ( &Rtp );

  

  MPI_Scatter( pAMatrix, Size  StripSize, MPI_DOUBLE, rbuf,  Size  StripSize, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  for (z = 0; z  ProcNum; z++)
  {
    if(ProcRank == 0)
    {
      for(i = 1; i  ProcNum; i++)
      {
        for (j = 0; j  StripSize; j++)
          MPI_Send(pBMatrix  + j + z  StripSize , 1, ColumnType, i , 0, MPI_COMM_WORLD);  
      }
    }
    else
    {

      double tmp;
      for (k = 0; k  StripSize; k++)
      {
        MPI_Recv(vertStrip, Size , MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);

        for(i = 0; i  StripSize; i++)
        {
          tmp = 0.0;
          for(j = 0; j  Size; j++)
          {
            tmp += rbuf[i  Size + j]  vertStrip[j]  ;
          }
          Result [i  StripSize + k] = tmp;
        }
      }
      MPI_Send( Result, StripSize  StripSize, MPI_DOUBLE , 0, 0, MPI_COMM_WORLD);
    }
    if(ProcRank == 0)
    {
      double tmp;
      for (i = 0; i  StripSize; i++)
      {

        for (j = 0; j  Size; j++)
        {
          tmp = 0.0;
          for(k = 0; k  Size; k++)
          {
            tmp += pAMatrix[ j  Size + k ]  pBMatrix[ k  Size + i + z  StripSize ]; 
          }
          pCMatrix[ j  Size + i + z  StripSize ] = tmp;
        }
      }
      for(i = 1; i  ProcNum; i++)
      {
        MPI_Recv(pCMatrix + iSizeStripSize + z  StripSize, 1, Rtp, i, 0, MPI_COMM_WORLD, &status);     
      }
    }
  }
  MPI_Type_free ( &ColumnType );
  MPI_Type_free ( &Rtp );
  free ( Result );
  free ( rbuf );
  free ( vertStrip );

  return 0;
}