﻿/*
// Реализация алгоритма параллельной сортировки 
// массива "пузырьком". 
//
// API
// int int OMP_BubbleSort ( double* pData, const int Size );
//
//  Input:
//
//     +  double* pData - исходный массив;
//     +  const int Size - размер массива;
//
//  Output:
//
//     +  double* pData - массив, отсортированный в порядке возрастания;
//
//  Result:
//     + функция возвращает значение типа int - код завершения функции.
*/

//int ThreadID;
//#pragma omp threadprivate(ThreadID) 
int OMP_BubbleSort(double* pData, const int Size)
{
  int *Indexes, *Blocks;
  int i;
  int ThreadID;
  int Iter;
  int ThreadNum;
  assert(pData);
  assert(Size>=7);

//	InitializeParallelSections();
  // инициализация параллельной секции
  #pragma omp parallel private (ThreadID)
	{
		ThreadID = omp_get_thread_num();
#pragma omp single
		{
			ThreadNum = omp_get_num_threads();
		}
	}
  // выделение памяти
	Indexes =(int*) malloc (ThreadNum * 2 * sizeof(int));
	Blocks = (int*) malloc (ThreadNum * 2 * sizeof(int));
	for (i=0; i < (2 * ThreadNum - 1); i++)
	{
		Indexes[i] = (int)((i * Size)/(double)(2 * ThreadNum));
    Blocks[i] = (int) (((i+1)*Size)/(double)(2*ThreadNum)) - Indexes[i];
	}
  Indexes[2 * ThreadNum - 1] = (int)(((2 * ThreadNum - 1) * Size)/(double)(2 * ThreadNum));
  Blocks [2 * ThreadNum - 1] = Size-Indexes[2 * ThreadNum - 1];

#pragma omp parallel private (ThreadID)
	{
    ThreadID = omp_get_thread_num();
		// Локальная сортировка
		QuickSorter(pData, Indexes[ThreadID*2], Indexes[ThreadID*2]+Blocks[2*ThreadID]+Blocks[2*ThreadID+1]-1);
    //SerialBubbleSort_b(pData, Indexes[ThreadID*2], Indexes[ThreadID*2]+Blocks[2*ThreadID]+Blocks[2*ThreadID+1]-1);
	}
#pragma omp barrier
	// Метод чет-нечетной перестановки
	Iter = 0;
	do {
		//printf("Iter\n");
#pragma omp parallel private (ThreadID)
    {
      ThreadID = omp_get_thread_num();
			// Четная итерация
			if (Iter%2 == 0)
			{
				if (ThreadID<ThreadNum-1)
				{
					MergeBlocks(pData, Indexes[2*ThreadID+1], Blocks[2*ThreadID+1], Indexes[2*ThreadID+2], Blocks[2*ThreadID+2]);
				}
			}
			//Нечетная итерация
			else
			{
					MergeBlocks(pData, Indexes[2*ThreadID], Blocks[2*ThreadID], Indexes[2*ThreadID+1], Blocks[2*ThreadID+1]);
			}
		}// pragma omp parallel
		Iter++;
	} 
  while (!IsSorted(pData, Size));
  free ( Indexes );
  free( Blocks );
  return 0;
}